"Turn templates into Cython pyx files."

import os.path

def template(funcs, bits, header):
    "Convert template dictionary `func` to a pyx file."
    codes = []
    for func in funcs: #supports multiple functions in a single file
        codes.append("# %s bit version\n" % str(bits))
        codes.append(func['main'])
        select = Selector(func['name'])
        for key in func['templates']:
            f = func['templates'][key]
            code = subtemplate(name=func['name'],
                               top=f['top'],
                               loop=f['loop'],
                               axisNone=f['axisNone'],
                               dtypes=f['dtypes'],
                               force_output_dtype=f['force_output_dtype'],
                               reuse_non_nan_func=f['reuse_non_nan_func'],
                               is_reducing_function=func['is_reducing_function'],
                               cdef_output=func['cdef_output'],
                               select=select,
                               bits=bits)
            codes.append(code)
            if 'sparse' in f:
                code = sparsetemplate(name=func['name'],
                                      template=f['sparse'],
                                      dtypes=f['dtypes'],
                                      select=select)
                codes.append(code)
        codes.append('\n' + str(select))
        if 'slow' in func:
            if func['slow'] is not None:
                slow = func['slow']
                code1 = slow_selector(slow['name'])
                code2 = slow_functions(slow['name'],
                                       slow['signature'],
                                       slow['func'])
                codes.append(code2)
                codes.append(code1)
    modpath = os.path.dirname(__file__)
    fid = open(os.path.join(modpath, '..', func['pyx_file']) % str(bits), 'w')
    fid.write(header)
    fid.write(''.join(codes))
    fid.close()


def sparsetemplate(name, template, dtypes, select):
    funcs = []
    for dtype in dtypes:
        func = template + "\n\n"
        func = func.replace('SPARSE', 'NAME_NDIMd_DTYPE_axisAXIS')
        func = func.replace('NAME', name)
        func = func.replace('DTYPE', dtype)
        func = func.replace('NDIM', '0')
        func = func.replace('AXIS', "None")
        funcs.append(func)
        select.append(0, dtype, None)
    return ''.join(funcs)

def subtemplate(name, top, loop, axisNone, dtypes, force_output_dtype,
                reuse_non_nan_func, is_reducing_function, cdef_output, select,
                bits):
    "Assemble template"
    ndims = sorted(loop.keys())
    funcs = []
    for ndim in ndims:
        if axisNone:
            axes = [None]
        else:
            axes = list(range(ndim))
        for dtype in dtypes:
            for axis in axes:

                if reuse_non_nan_func:

                    select.append(ndim, dtype, axis, True)

                else:

                    # Code template
                    func = top

                    # loop
                    if force_output_dtype is not False:
                        ydtype = force_output_dtype
                    else:
                        ydtype = dtype
                    func += loop_cdef(ndim, ydtype, axis, is_reducing_function,
                                      cdef_output)
                    func += looper(loop[ndim], ndim, axis)

                    # name, ndim, dtype, axis
                    func = func.replace('NAME', name)
                    func = func.replace('NDIM', str(ndim))
                    func = func.replace('DTYPE', dtype)
                    func = func.replace('AXIS', str(axis))

                    funcs.append(func)
                    select.append(ndim, dtype, axis)

    return ''.join(funcs)

def looper(loop, ndim, axis):
    """
    Given loop template, expand index markers for given `ndim` and `axis`.

    Parameters
    ----------
    loop : str
        Code of loop where the following template markers will be expanded
        (example given is for 3d input, similarly for other nd):

        ================= =================================================
        INDEXALL          Replace with i0, i1, i2
        INDEXPOP          If axis=1, e.g., replace with i0, i2
        INDEXN            If N=1, e.g., replace with 1
        INDEXREPLACE|exp| If exp = 'k - window' and axis=1, e.g., replace
                          with i0, k - window, i2
        NREPLACE|exp|     If exp = 'n - window' and axis=1, e.g., replace
                          with n0, n - window, n2
        ================= =================================================
    ndim : int
        Number of dimensions in the loop.
    axis : {int, None}
        Axis over which the loop is evaluated.

    Returns
    -------
    code : str
        Code for the loop with templated index markers expanded.

    Examples
    --------
    Make a 3d loop template:

    >>> loop = '''
    .... for iINDEX0 in range(nINDEX0):
    ....    for iINDEX1 in range(nINDEX1):
    ....        amin = MAXDTYPE
    ....        for iINDEX2 in range(nINDEX2):
    ....            ai = a[INDEXALL]
    ....            if ai <= amin:
    ....                amin = ai
    ....         y[INDEXPOP] = amin
    .... '''

    Import the looper function:

    >>> from bottlechest.src.template.template import looper

    Make a loop over axis=0:

    >>> print(looper(loop, ndim=3, axis=0))
    for i1 in range(n1):
        for i2 in range(n2):
            amin = MAXDTYPE
            for i0 in range(n0):
                ai = a[i0, i1, i2]
                if ai <= amin:
                    amin = ai
            y[i1, i2] = amin

    Make a loop over axis=1:

    >>> print(looper(loop, ndim=3, axis=1))
    for i0 in range(n0):
        for i2 in range(n2):
            amin = MAXDTYPE
            for i1 in range(n1):
                ai = a[i0, i1, i2]
                if ai <= amin:
                    amin = ai
            y[i0, i2] = amin

    Make a loop over axis=2:

    >>> print(looper(loop, ndim=3, axis=2))
    for i0 in range(n0):
        for i1 in range(n1):
            amin = MAXDTYPE
            for i2 in range(n2):
                ai = a[i0, i1, i2]
                if ai <= amin:
                    amin = ai
            y[i0, i1] = amin

    """

    if ndim < 1:
        raise ValueError("ndim(=%d) must be and integer greater than 0" % ndim)
    if axis is not None:
        if axis < 0:
            raise ValueError("`axis` must be a non-negative integer or None")
        elif axis >= ndim:
            raise ValueError("`axis` must be less then `ndim`")

    # INDEXALL
    INDEXALL = ', '.join('i' + str(i) for i in range(ndim))
    code = loop.replace('INDEXALL', INDEXALL)

    # INDEXPOP
    idx = list(range(ndim))
    if axis is not None:
        idx.pop(axis)
    INDEXPOP = ', '.join(['i' + str(i) for i in idx])
    code = code.replace('INDEXPOP', INDEXPOP)

    # INDEXN
    idx = list(range(ndim))
    if axis is not None:
        idxpop = idx.pop(axis)
        idx.append(idxpop)
    for i, j in enumerate(idx):
        code = code.replace('INDEX%d' % i, '%d' % j)

    # INDEXREPLACE|x|
    mark = 'INDEXREPLACE|'
    nreplace = code.count(mark)
    if (nreplace > 0) and (axis is None):
        raise ValueError("`INDEXREPLACE` cannot be used when axis is None.")
    while mark in code:
        idx0 = code.index(mark)
        idx1 = idx0 + len(mark)
        idx2 = idx1 + code[idx1:].index('|')
        if (idx0 >= idx1) or (idx1 >= idx2):
            raise RuntimeError("Parsing error or poorly formatted input.")
        replacement = code[idx1:idx2]
        idx = ['i' + str(i) for i in range(ndim)]
        idx[axis] = replacement
        idx = ', '.join(idx)
        code = code[:idx0] + idx + code[idx2+1:]

    # NREPLACE|x|
    mark = 'NREPLACE|'
    nreplace = code.count(mark)
    # TODO: reuse while loop above, only difference is 'i' --> 'n'
    while mark in code:
        idx0 = code.index(mark)
        idx1 = idx0 + len(mark)
        idx2 = idx1 + code[idx1:].index('|')
        if (idx0 >= idx1) or (idx1 >= idx2):
            raise RuntimeError("Parsing error or poorly formatted input.")
        replacement = code[idx1:idx2]
        idx = ['n' + str(i) for i in range(ndim)]
        idx[axis] = replacement
        idx = ', '.join(idx)
        code = code[:idx0] + idx + code[idx2+1:]

    return code

def loop_cdef(ndim, dtype, axis, is_reducing_function, cdef_output=True):
    """
    String of code that initializes variables needed in a for loop.

    The output string contains code for: index array counters, one for each
    dimension (cdef Py_size_t i0, i1, i2, ....); the length along each
    dimension of the input array, `a` (cdef Py_ssize_t n0 = a.shape[0],...);
    the initialized, empty output array, `y`.

    Parameters
    ----------
    ndim = int
        Number of dimensions.
    dtype : str
        The data type of the output. Used for initilizing the empty output
        array, `y`.
    axis : {int, None}
        If `is_reducing_function` is True then remove the dimension given
        by `axis` when initializing the output array, `y`.
    is_reducing_function : bool
        If True then remove the dimension given by `axis` when initializing
        the output array, `y`.
    cdef_output : bool, optional
        If False then only initialize indices (i) and shapes (n). If True
        (default) then also intialized output array `y`.

    Returns
    -------
    cdefs : str
        String of code to use to initialize variables needed for loop.

    Examples
    --------
    Define parameters:

    >>> ndim = 3
    >>> dtype = 'float64'
    >>> axis = 1
    >>> is_reducing_function = True

    Import loop_cdef:

    >>> from bottlechest.src.template.template import loop_cdef

    Make loop initialization code:

    >>> print(loop_cdef(ndim, dtype, axis, is_reducing_function))
        cdef Py_ssize_t i0, i1, i2
        cdef np.npy_intp *dim
        dim = PyArray_DIMS(a)
        Py_ssize_t n0 = dim[0]
        Py_ssize_t n1 = dim[1]
        Py_ssize_t n2 = dim[2]
        cdef np.npy_intp *dims = [n0, n2]
        cdef np.ndarray[np.float64_t, ndim=2] y = PyArray_EMPTY(2, dims,
                                                  NPY_float64, 0)

    Repeat, but this time make the output non-reducing:

    >>> is_reducing_function = False
    >>> print(loop_cdef(ndim, dtype, axis, is_reducing_function))
        cdef Py_ssize_t i0, i1, i2
        cdef np.npy_intp *dim
        dim = PyArray_DIMS(a)
        Py_ssize_t n0 = dim[0]
        Py_ssize_t n1 = dim[1]
        Py_ssize_t n2 = dim[2]
        cdef np.npy_intp *dims = [n0, n1, n2]
        cdef np.ndarray[np.float64_t, ndim=3] y = PyArray_EMPTY(3, dims,
                                                  NPY_float64, 0)

    """

    if ndim < 1:
        raise ValueError("ndim(=%d) must be and integer greater than 0" % ndim)
    if axis is not None:
        if axis < 0:
            raise ValueError("`axis` must be a non-negative integer or None")
        elif axis >= ndim:
            raise ValueError("`axis` must be less then `ndim`")

    tab = '    '
    cdefs = []

    # cdef loop indices
    idx = ', '.join('i'+str(i) for i in range(ndim))
    cdefs.append(tab + 'cdef Py_ssize_t ' + idx)

    # Length along each dimension
    cdefs.append(tab + "cdef np.npy_intp *dim")
    cdefs.append(tab + "dim = PyArray_DIMS(a)")
    for dim in range(ndim):
        cdefs.append(tab + "cdef Py_ssize_t n%d = dim[%d]" % (dim, dim))

    if not cdef_output:
        return '\n'.join(cdefs) + '\n'

    # cdef initialize output
    if is_reducing_function:
        if (ndim > 1) and (axis is not None):
            idx = list(range(ndim))
            del idx[axis]
            ns = ', '.join(['n'+str(i) for i in idx])
            cdefs.append("%scdef np.npy_intp *dims = [%s]" % (tab, ns))
            if dtype == 'bool':
                y = "%scdef np.ndarray[np.uint8_t, ndim=%d, cast=True] "
                y += "y = PyArray_EMPTY(%d, dims,"
                y += "\n\t\tNPY_BOOL, 0)"
                cdefs.append(y % (tab, ndim-1, ndim-1))
            else:
                y = "%scdef np.ndarray[np.%s_t, ndim=%d] "
                y += "y = PyArray_EMPTY(%d, dims,"
                y += "\n\t\tNPY_%s, 0)"
                cdefs.append(y % (tab, dtype, ndim-1, ndim-1, dtype))
    else:
        idx = list(range(ndim))
        ns = ', '.join('n'+str(i) for i in idx)
        cdefs.append("%scdef np.npy_intp *dims = [%s]" % (tab, ns))
        if dtype == 'bool':
            y = "%scdef np.ndarray[np.uint8_t, ndim=%d, cast=True] "
            y += "y = PyArray_EMPTY(%d, dims,"
            y += "\n\t\tNPY_BOOL, 0)"
            cdefs.append(y % (tab, ndim, ndim))
        else:
            y = "%scdef np.ndarray[np.%s_t, ndim=%d] "
            y += "y = PyArray_EMPTY(%d, dims,"
            y += "\n\t\tNPY_%s, 0)"
            cdefs.append(y % (tab, dtype, ndim, ndim, dtype))

    return '\n'.join(cdefs) + '\n'

class Selector(object):
    "String of code for dictionary that maps dtype to cython function."

    def __init__(self, name):
        self.name = name
        self.data = []

    def append(self, ndim, dtype, axis, reuse=False):
        self.data.append((ndim, dtype, axis, reuse))

    def __str__(self):
        fmt = "%s_dict[(%s, NPY_%s, %s)] = %s_%sd_%s_axis%s"
        src = []
        src.append("cdef dict %s_dict = {}" % self.name)
        for ndim, dtype, axis, reuse in self.data:
            name = self.name
            if reuse:
                name = name.replace('nan', '')
            if (ndim == 1) and (axis is None):
                tup = (self.name, str(ndim), str(dtype), str(0),
                       name, str(ndim), str(dtype), str(axis))
                src.append(fmt % tup)
            tup = (self.name, str(ndim), str(dtype), str(axis),
                   name, str(ndim), str(dtype), str(axis))
            src.append(fmt % tup)
        return '\n'.join(src)

def slow_selector(name, maxaxis=32):
    "String of code for slow function mapping dictionary."
    axes = list(range(maxaxis+1)) + [None]
    src = ['\n']
    src.append("cdef dict %s_slow_dict = {}" % name)
    fmt = "%s_slow_dict[%s] = %s_slow_axis%s"
    for axis in axes:
        tup = 2 * (name, str(axis))
        src.append(fmt % tup)
    return '\n'.join(src)

def slow_functions(name, signature, func, maxaxis=32):
    "String of code for slow functions."
    axes = list(range(maxaxis+1)) + [None]
    tab = '    '
    sig = "def %s_slow_axis%s(%s):"
    doc = '%s"Unaccelerated (slow) %s along axis %s."'
    function = "%sreturn %s\n"
    src = ['\n']
    for axis in axes:

        axis = str(axis)

        # signature
        code = sig % (name, axis, signature)
        code = code.replace('AXIS', axis)
        src.append(code)

        # docstring
        code = doc % (tab, name, axis)
        code = code.replace('AXIS', axis)
        src.append(code)

        # function
        code = function % (tab, func)
        code = code.replace('AXIS', axis)
        src.append(code)

    return '\n'.join(src)



import unittest
import sys
import bottlechest as bn
import numpy as np
import scipy.sparse as sp


class TestBinCount(unittest.TestCase):
    def test_1d_int(self):
        data = np.array([0, 1, 1, 2, 1])
        count, nans = bn.bincount(data, 2)
        np.testing.assert_equal(count, [1, 3, 1])
        self.assertEqual(nans, 0)

    def test_1d_float(self):
        nan = float("nan")
        data = np.array([0, nan, nan, 2, 2, 1], dtype=float)
        count, nans = bn.bincount(data, 2)
        np.testing.assert_equal(count, [1, 1, 2])
        self.assertEqual(nans, 2)

    def test_1d_mask_int(self):
        data = np.array([0, 1, 1, 2, 1])
        count, nans = bn.bincount(data, 2, mask=[1])
        np.testing.assert_equal(count, [1, 3, 1])
        self.assertEqual(nans, 0)

        count, nans = bn.bincount(data, 2, mask=[0])
        np.testing.assert_equal(count, [0, 0, 0])
        self.assertEqual(nans, 0)

    def test_1d_mask_float(self):
        nan = float("nan")
        data = np.array([0, nan, nan, 2, 2, 1], dtype=float)
        count, nans = bn.bincount(data, 2, mask=[1])
        np.testing.assert_equal(count, [1, 1, 2])
        self.assertEqual(nans, 2)

        count, nans = bn.bincount(data, 2, mask=[0])
        np.testing.assert_equal(count, [0, 0, 0])
        self.assertEqual(nans, 0)

    def test_1d_weighted_int(self):
        data = np.array([0, 1, 1, 2, 1])
        count, nans = bn.bincount(data, 2, weights=np.arange(5))
        np.testing.assert_equal(count, [0, 7, 3])
        self.assertEqual(nans, 0)

    def test_1d_weighted_float(self):
        nan = float("nan")
        data = np.array([0, nan, nan, 2, 2, 1], dtype=float)
        count, nans = bn.bincount(data, 2, weights=np.arange(6))
        np.testing.assert_equal(count, [0, 5, 7])
        self.assertEqual(nans, 3)

    def test_simple_int(self):
        data = np.array([[0, 1, 1, 2, 1],
                         [1, 1, 1, 0, 1],
                         [0, 0, 3, 0, 0]], dtype=int)
        counts, nans = bn.bincount(data, 3)
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 2, 0, 1], [2, 0, 1, 0], [1, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 0])

    def test_simple_float(self):
        nan = float("nan")
        data = np.array([[0, 1, 1, 2, 1],
                         [1, 1, 1, nan, 1],
                         [0, 0, 3, nan, nan]], dtype=float)
        counts, nans = bn.bincount(data, 3)
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 2, 0, 1], [0, 0, 1, 0], [0, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 2, 1])

    def test_weighted_int(self):
        data = np.array([[0, 1, 1, 2, 1],
                         [1, 1, 1, 0, 1],
                         [0, 0, 3, 0, 0]], dtype=float)
        counts, nans = bn.bincount(data, 3, np.array([1, 2, 3], dtype=int))
        print(counts)
        np.testing.assert_almost_equal(counts, [[4, 2, 0, 0], [3, 3, 0, 0], [0, 3, 0, 3], [5, 0, 1, 0], [3, 3, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 0])

    def test_weighted_float(self):
        nan = float("nan")
        data = np.array([[0, 1, 1, 2, 1],
                         [1, 1, 1, nan, 1],
                         [0, 0, 3, nan, nan]], dtype=float)
        counts, nans = bn.bincount(data, 3, [1, 2, 3])
        np.testing.assert_almost_equal(counts, [[4, 2, 0, 0], [3, 3, 0, 0], [0, 3, 0, 3], [0, 0, 1, 0], [0, 3, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 5, 3])

    def test_mask_int(self):
        data = np.array([[0, 1, 1, 2, 1],
                         [1, 1, 1, 0, 1],
                         [0, 0, 3, 0, 0]], dtype=float)
        counts, nans = bn.bincount(data, 3, mask=np.array([1, 1, 0, 0, 1]))
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [1, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 0])
        counts, nans = bn.bincount(data, 3, mask=np.array([1, 1, 0, 0, 1], dtype=np.int8))
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [1, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 0])
        counts, nans = bn.bincount(data, 3, mask=np.array([1, 1, 0, 0, 1], dtype=np.int32))
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [1, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 0])
        counts, nans = bn.bincount(data, 3, mask=np.array([1, 1, 0, 0, 1])==1)
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [1, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 0])

    def test_mask_float(self):
        nan = float("nan")
        data = np.array([[0, 1, 1, 2, 1],
                         [1, 1, 1, nan, 1],
                         [0, 0, 3, nan, nan]], dtype=float)
        counts, nans = bn.bincount(data, 3, mask=np.array([1, 1, 0, 0, 1]))
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 1])
        counts, nans = bn.bincount(data, 3, mask=np.array([1, 1, 0, 0, 1], dtype=np.int8))
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 1])
        counts, nans = bn.bincount(data, 3, mask=np.array([1, 1, 0, 0, 1], dtype=np.int8))
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 1])
        counts, nans = bn.bincount(data, 3, mask=np.array([1, 1, 0, 0, 1])==1)
        np.testing.assert_almost_equal(counts, [[2, 1, 0, 0], [1, 2, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 2, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 1])

    def test_weighted_mask_int(self):
        data = np.array([[0, 1, 1, 2, 1],
                         [1, 1, 1, 0, 1],
                         [0, 0, 3, 0, 0]], dtype=float)
        counts, nans = bn.bincount(data, 3, weights=[1, 2, 3], mask=np.array([1, 1, 0, 0, 1]))
        np.testing.assert_almost_equal(counts, [[4, 2, 0, 0], [3, 3, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [3, 3, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 0])

    def test_weighted_mask_float(self):
        nan = float("nan")
        data = np.array([[0, 1, 1, 2, 1],
                         [1, 1, 1, nan, 1],
                         [0, 0, 3, nan, nan]], dtype=float)
        counts, nans = bn.bincount(data, 3, weights=[1, 2, 3], mask=np.array([1, 1, 0, 0, 1]))
        np.testing.assert_almost_equal(counts, [[4, 2, 0, 0], [3, 3, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 3, 0, 0]])
        np.testing.assert_almost_equal(nans, [0, 0, 0, 0, 3])

    def test_sparse_int(self):
        data = np.array([1, 1, 2, 2, 1, 3])
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        counts, nans = bn.bincount(a, 3)
        np.testing.assert_almost_equal(counts, [[0, 1, 1, 0], [0, 2, 0, 0], [0, 0, 1, 1], [0, 0, 0, 0]])

    def test_sparse_float(self):
        data = np.array([1, 1, 2, 2, 1, 3], dtype=float)
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        counts, nans = bn.bincount(a, 3)
        np.testing.assert_almost_equal(counts, [[0, 1, 1, 0], [0, 2, 0, 0], [0, 0, 1, 1], [0, 0, 0, 0]])


    def test_sparse_weight_int(self):
        data = np.array([1, 1, 2, 2, 1, 3])
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        counts, nans = bn.bincount(a, 3, weights=[1, 2, 3])
        np.testing.assert_almost_equal(counts, [[0, 1, 2, 0], [0, 4, 0, 0], [0, 0, 1, 3], [0, 0, 0, 0]])

    def test_sparse_weight_float(self):
        data = np.array([1, 1, 2, 2, 1, 3], dtype=float)
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        counts, nans = bn.bincount(a, 3, np.array([1, 2, 3], dtype=float))
        np.testing.assert_almost_equal(counts, [[0, 1, 2, 0], [0, 4, 0, 0], [0, 0, 1, 3], [0, 0, 0, 0]])


    def test_sparse_mask_int(self):
        data = np.array([1, 1, 2, 2, 1, 3])
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        counts, nans = bn.bincount(a, 3, mask=[1, 0, 1, 0])
        np.testing.assert_almost_equal(counts, [[0, 1, 1, 0], [0, 0, 0, 0], [0, 0, 1, 1], [0, 0, 0, 0]])

    def test_sparse_mask_float(self):
        data = np.array([1, 1, 2, 2, 1, 3], dtype=float)
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        counts, nans = bn.bincount(a, 3, mask=[1, 0, 1, 0])
        np.testing.assert_almost_equal(counts, [[0, 1, 1, 0], [0, 0, 0, 0], [0, 0, 1, 1], [0, 0, 0, 0]])


    def test_sparse_weight_int(self):
        data = np.array([1, 1, 2, 2, 1, 3])
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        counts, nans = bn.bincount(a, 3, weights=[1, 2, 3], mask=[1, 0, 1, 0])
        np.testing.assert_almost_equal(counts, [[0, 1, 2, 0], [0, 0, 0, 0], [0, 0, 1, 3], [0, 0, 0, 0]])

    def test_sparse_weight_float(self):
        data = np.array([1, 1, 2, 2, 1, 3], dtype=float)
        indptr = [0, 3, 4, 6]
        indices = [0, 1, 2, 0, 1, 2]
        a = sp.csr_matrix((data, indices, indptr), shape=(3, 4))
        counts, nans = bn.bincount(a, 3, np.array([1, 2, 3], dtype=float), mask=[1, 0, 1, 0])
        np.testing.assert_almost_equal(counts, [[0, 1, 2, 0], [0, 0, 0, 0], [0, 0, 1, 3], [0, 0, 0, 0]])





"nanmean template"

from copy import deepcopy
import bottlechest as bn

__all__ = ["nanmean"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = False
floats['force_output_dtype'] = False
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a):
    "Mean of NDIMd array with dtype=DTYPE along axis=AXIS ignoring NaNs."
    cdef Py_ssize_t count = 0
    cdef np.DTYPE_t asum = 0, ai
"""

loop = {}
loop[2] = """\
    for iINDEX0 in range(nINDEX0):
        asum = 0
        count = 0
        for iINDEX1 in range(nINDEX1):
            ai = a[INDEXALL]
            if ai == ai:
                asum += ai
                count += 1
        if count > 0:       
            y[INDEXPOP] = asum / count
        else:
            y[INDEXPOP] = NAN
    return y
"""
floats['loop'] = loop

# Float dtypes (axis=None) --------------------------------------------------

floats_None = deepcopy(floats)
floats_None['axisNone'] = True

returns = """\
    if count > 0:
        return np.DTYPE(asum / count)
    else:
        return np.DTYPE(NAN)
"""        

loop = {}
loop[1] = """\
    for iINDEX0 in range(nINDEX0):
        ai = a[INDEXALL]
        if ai == ai:
            asum += ai
            count += 1
""" + returns
loop[2] = """\
    for iINDEX0 in range(nINDEX0):
        for iINDEX1 in range(nINDEX1):
            ai = a[INDEXALL]
            if ai == ai:
                asum += ai
                count += 1
""" + returns
floats_None['loop'] = loop

# Int dtypes (not axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES 
ints['force_output_dtype'] = 'float64'

ints['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a):
    "Mean of NDIMd array with dtype=DTYPE along axis=AXIS ignoring NaNs."
    cdef np.float64_t asum = 0, ai
"""

loop = {}
loop[2] = """\
    if nINDEX1 == 0:
        PyArray_FillWithScalar(y, NAN)
    else:
        for iINDEX0 in range(nINDEX0):
            asum = 0
            for iINDEX1 in range(nINDEX1):
                asum += a[INDEXALL]
            y[INDEXPOP] = asum / nINDEX1
    return y
"""
ints['loop'] = loop

# Int dtypes (axis=None) ----------------------------------------------------

ints_None = deepcopy(ints) 
ints_None['top'] = ints['top'] + "    cdef Py_ssize_t size\n"
ints_None['axisNone'] = True

loop = {}
loop[1] = """\
    size = nINDEX0
    for iINDEX0 in range(nINDEX0):
        asum += a[INDEXALL]
    if size > 0:    
        return np.float64(asum / size)
    else:
        return np.float64(NAN)
"""
loop[2] = """\
    size = nINDEX0 * nINDEX1    
    for iINDEX0 in range(nINDEX0):
        for iINDEX1 in range(nINDEX1):
            asum += a[INDEXALL]
    if size > 0:    
        return np.float64(asum / size)
    else:
        return np.float64(NAN)
"""
ints_None['loop'] = loop

# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "nanmean"
slow['signature'] = "arr"
slow['func'] = "bn.slow.nanmean(arr, axis=AXIS)"

# Template ------------------------------------------------------------------

nanmean = {}
nanmean['name'] = 'nanmean'
nanmean['is_reducing_function'] = True
nanmean['cdef_output'] = True
nanmean['slow'] = slow
nanmean['templates'] = {}
nanmean['templates']['float'] = floats
nanmean['templates']['float_None'] = floats_None
nanmean['templates']['int'] = ints
nanmean['templates']['int_None'] = ints_None
nanmean['pyx_file'] = 'func/%sbit/nanmean.pyx'

nanmean['main'] = '''"nanmean auto-generated from template"

def nanmean(arr, axis=None):
    """
    Mean of array elements along given axis ignoring NaNs.

    `float64` intermediate and return values are used for integer inputs.

    Parameters
    ----------
    arr : array_like
        Array containing numbers whose mean is desired. If `arr` is not an
        array, a conversion is attempted.
    axis : {int, None}, optional
        Axis along which the mean is computed. The default (axis=None) is to
        compute the mean of the flattened array.

    Returns
    -------
    y : ndarray
        An array with the same shape as `arr`, with the specified axis removed.
        If `arr` is a 0-d array, or if axis is None, a scalar is returned.
        `float64` intermediate and return values are used for integer inputs. 

    See also
    --------
    bottlechest.nanmedian: Median along specified axis, ignoring NaNs.
    
    Notes
    -----
    No error is raised on overflow. (The sum is computed and then the result
    is divided by the number of non-NaN elements.)

    If positive or negative infinity are present the result is positive or
    negative infinity. But if both positive and negative infinity are present,
    the result is Not A Number (NaN).

    Examples
    --------
    >>> bn.nanmean(1)
    1.0
    >>> bn.nanmean([1])
    1.0
    >>> bn.nanmean([1, np.nan])
    1.0
    >>> a = np.array([[1, 4], [1, np.nan]])
    >>> bn.nanmean(a)
    2.0
    >>> bn.nanmean(a, axis=0)
    array([ 1.,  4.])

    When positive infinity and negative infinity are present:

    >>> bn.nanmean([1, np.nan, np.inf])
    inf
    >>> bn.nanmean([1, np.nan, np.NINF])
    -inf
    >>> bn.nanmean([1, np.nan, np.inf, np.NINF])
    nan
    
    """
    func, arr = nanmean_selector(arr, axis)
    return func(arr)

def nanmean_selector(arr, axis):
    """
    Return nanmean function and array that matches `arr` and `axis`.
    
    Under the hood Bottleneck uses a separate Cython function for each
    combination of ndim, dtype, and axis. A lot of the overhead in
    bn.nanmean() is in checking that `axis` is within range, converting `arr`
    into an array (if it is not already an array), and selecting the function
    to use to calculate the mean.

    You can get rid of the overhead by doing all this before you, for example,
    enter an inner loop, by using the this function.

    Parameters
    ----------
    arr : array_like
        Input array. If `arr` is not an array, a conversion is attempted.
    axis : {int, None}
        Axis along which the mean is to be computed.
    
    Returns
    -------
    func : function
        The nanmean function that matches the number of dimensions and dtype
        of the input array and the axis along which you wish to find the mean.
    a : ndarray
        If the input array `arr` is not a ndarray, then `a` will contain the
        result of converting `arr` into a ndarray.

    Examples
    --------
    Create a numpy array:

    >>> arr = np.array([1.0, 2.0, 3.0])
    
    Obtain the function needed to determine the nanmean of `arr` along axis=0:

    >>> func, a = bn.func.nanmean_selector(arr, axis=0)
    >>> func
    <function nanmean_1d_float64_axis0>
    
    Use the returned function and array to determine the mean:

    >>> func(a)
    2.0

    """
    cdef np.ndarray a
    if type(arr) is np.ndarray:
        a = arr
    else:    
        a = np.array(arr, copy=False)
    cdef int ndim = PyArray_NDIM(a)
    cdef int dtype = PyArray_TYPE(a)
    if (axis is not None) and (axis < 0):
        axis += ndim
    cdef tuple key = (ndim, dtype, axis)
    try:
        func = nanmean_dict[key]
    except KeyError:
        if axis is not None:
            if (axis < 0) or (axis >= ndim):
                raise ValueError("axis(=%d) out of bounds" % axis)
        try:
            func = nanmean_slow_dict[axis]
        except KeyError:
            tup = (str(ndim), str(a.dtype), str(axis))
            raise TypeError("Unsupported ndim/dtype/axis (%s/%s/%s)." % tup)
    return func, a
'''   



"""Generate pyx files from templates.

To generate a single file, run this script with the argument.
For example, use "python makepyx.py func.median"
"""

from bottlechest.src.template.package import (package, single)

def makepyx(): #makes all pyxfiles
    package("func", bits=32)
    package("func", bits=64)

def makepyxsingle(name):
    single(name, bits=32)
    single(name, bits=64)

if __name__ == "__main__":
    import sys
    makepyxsingle(sys.argv[1])
    



"anynan template"

from copy import deepcopy
import bottlechest as bn

__all__ = ["anynan"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = False
floats['force_output_dtype'] = 'bool'
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a):
    "Check for NaNs in NDIMd array with dtype=DTYPE along axis=AXIS."
    cdef int f = 1
    cdef np.DTYPE_t ai
"""

loop = {}
loop[2] = """\
    for iINDEX0 in range(nINDEX0):
        f = 1
        for iINDEX1 in range(nINDEX1):
            ai = a[INDEXALL]
            if ai != ai:
                y[INDEXPOP] = 1
                f = 0
                break
        if f == 1:
            y[INDEXPOP] = 0
    return y
"""

floats['loop'] = loop

# Float dtypes (axis=None) --------------------------------------------------

floats_None = deepcopy(floats)
floats_None['axisNone'] = True

loop = {}
loop[1] = """\
    for iINDEX0 in range(nINDEX0):
        ai = a[INDEXALL]
        if ai != ai:
            return np.bool_(True)
    return np.bool_(False)
"""
loop[2] = """\
    for iINDEX0 in range(nINDEX0):
        for iINDEX1 in range(nINDEX1):
            ai = a[INDEXALL]
            if ai != ai:
                return np.bool_(True)
    return np.bool_(False)
"""

floats_None['loop'] = loop

# Int dtypes (not axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES 

loop = {}
loop[2] = """\
    for iINDEX0 in range(nINDEX0):
        y[INDEXPOP] = 0
    return y
"""

ints['loop'] = loop

# Int dtypes (axis=None) ----------------------------------------------------

ints_None = deepcopy(ints) 
ints_None['axisNone'] = True

loop = {}
loop[1] = """\
    return np.bool_(False)
"""
loop[2] = """\
    return np.bool_(False)
"""

ints_None['loop'] = loop

# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "anynan"
slow['signature'] = "arr"
slow['func'] = "bn.slow.anynan(arr, axis=AXIS)"

# Template ------------------------------------------------------------------

anynan = {}
anynan['name'] = 'anynan'
anynan['is_reducing_function'] = True
anynan['cdef_output'] = True
anynan['slow'] = slow
anynan['templates'] = {}
anynan['templates']['float'] = floats
anynan['templates']['float_None'] = floats_None
anynan['templates']['int'] = ints
anynan['templates']['int_None'] = ints_None
anynan['pyx_file'] = 'func/%sbit/anynan.pyx'

anynan['main'] = '''"anynan auto-generated from template"

def anynan(arr, axis=None):
    """
    Test whether any array element along a given axis is NaN.

    Returns single boolean unless `axis` is not ``None``.

    Parameters
    ----------
    arr : array_like
        Input array. If `arr` is not an array, a conversion is attempted.
    axis : {int, None}, optional
        Axis along which NaNs are searched.  The default (`axis` = ``None``)
        is to search for NaNs over a flattened input array. `axis` may be
        negative, in which case it counts from the last to the first axis.

    Returns
    -------
    y : bool or ndarray
        A new boolean or `ndarray` is returned.

    See also
    --------
    bottlechest.allnan: Test if all array elements along given axis are NaN

    Examples
    --------
    >>> bn.anynan(1)
    False
    >>> bn.anynan(np.nan)
    True
    >>> bn.anynan([1, np.nan])
    True
    >>> a = np.array([[1, 4], [1, np.nan]])
    >>> bn.anynan(a)
    True
    >>> bn.anynan(a, axis=0)
    array([False,  True], dtype=bool)    

    """
    func, arr = anynan_selector(arr, axis)
    return func(arr)

def anynan_selector(arr, axis):
    """
    Return anynan function and array that matches `arr` and `axis`.
    
    Under the hood Bottleneck uses a separate Cython function for each
    combination of ndim, dtype, and axis. A lot of the overhead in bn.anynan()
    is in checking that `axis` is within range, converting `arr` into an
    array (if it is not already an array), and selecting the function to use.

    You can get rid of the overhead by doing all this before you, for example,
    enter an inner loop, by using the this function.

    Parameters
    ----------
    arr : array_like
        Input array. If `arr` is not an array, a conversion is attempted.
    axis : {int, None}
        Axis along which NaNs are searched.
    
    Returns
    -------
    func : function
        The anynan function that matches the number of dimensions and
        dtype of the input array and the axis.
    a : ndarray
        If the input array `arr` is not a ndarray, then `a` will contain the
        result of converting `arr` into a ndarray.

    Examples
    --------
    Create a numpy array:

    >>> arr = np.array([1.0, 2.0, 3.0])
    
    Obtain the function needed to determine if there are any NaN in `arr`:

    >>> func, a = bn.func.anynan_selector(arr, axis=0)
    >>> func
    <function anynan_1d_float64_axisNone>
    
    Use the returned function and array to determine if there are any
    NaNs:
    
    >>> func(a)
    False

    """
    cdef np.ndarray a
    if type(arr) is np.ndarray:
        a = arr
    else:    
        a = np.array(arr, copy=False)
    cdef int ndim = PyArray_NDIM(a)
    cdef int dtype = PyArray_TYPE(a)
    if (axis is not None) and (axis < 0):
        axis += ndim
    cdef tuple key = (ndim, dtype, axis)
    try:
        func = anynan_dict[key]
    except KeyError:
        if axis is not None:
            if (axis < 0) or (axis >= ndim):
                raise ValueError("axis(=%d) out of bounds" % axis)
        try:
            func = anynan_slow_dict[axis]
        except KeyError:
            tup = (str(ndim), str(a.dtype), str(axis))
            raise TypeError("Unsupported ndim/dtype/axis (%s/%s/%s)." % tup)
    return func, a
'''   



from distutils.core import setup
import setuptools
import sys, os

version = "0.4"

setup(
    name="numpydoc",
    packages=["numpydoc"],
    package_dir={"numpydoc": ""},
    version=version,
    description="Sphinx extension to support docstrings in Numpy format",
    # classifiers from http://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=["Development Status :: 3 - Alpha",
                 "Environment :: Plugins",
                 "License :: OSI Approved :: BSD License",
                 "Topic :: Documentation"],
    keywords="sphinx numpy",
    author="Pauli Virtanen and others",
    author_email="pav@iki.fi",
    url="http://github.com/numpy/numpy/tree/master/doc/sphinxext",
    license="BSD",
    zip_safe=False,
    install_requires=["Sphinx >= 1.0.1"],
    package_data={'numpydoc': 'tests', '': ''},
    entry_points={
        "console_scripts": [
            "autosummary_generate = numpydoc.autosummary_generate:main",
        ],
    },
)



"bincount template"

from copy import deepcopy
import bottlechest as bn

__all__ = ["bincount"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = True
floats['force_output_dtype'] = 'bool'
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a,
                              int max_val,
                              np.ndarray[np.float_t, ndim=1] w = None,
                              np.ndarray[np.int8_t, ndim=1] mask = None):
    '''Bincount that can handle floats (casts to int), handles NaNs and needs to
     be specified a fixed number of values'''
"""

loop = {}

loop[1] = """\
    if w is not None and len(w) != n0:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.npy_intp *dims = [max_val+1]
        np.ndarray[np.float64_t, ndim=1] y = PyArray_ZEROS(1, dims, NPY_float64, 0)
        float nans = 0
        int ain
        float ai, wt
    if mask is not None and not mask[0]:
        return y, nans
    if w is None:
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai != ai:
                nans += 1
                continue
            if ai < -1e-6:
                raise ValueError("negative value in bincount")
            ain = int(ai + 0.1)
            if not -1e-6 < ain - ai < 1e-6:
                raise ValueError("%f is not an integer value" % ai)
            if ain > max_val:
                raise ValueError("value %i is greater than max_val (%i)" %
                                 (ain, max_val))
            y[ain] += 1
    else:
        for iINDEX0 in range(nINDEX0):
            wt = w[iINDEX0]
            ai = a[INDEXALL]
            if ai != ai:
                nans += wt
                continue
            if ai < -1e-6:
                raise ValueError("negative value in bincount")
            ain = int(ai + 0.1)
            if not -1e-6 < ain - ai < 1e-6:
                raise ValueError("%f is not an integer value" % ai)
            if ain > max_val:
                raise ValueError("value %i is greater than max_val (%i)" %
                                 (ain, max_val))
            y[ain] += wt
    return y, nans
"""

loop[2] = """\
    if w is not None and len(w) != len(a):
        raise ValueError("invalid length of the weight vector")

    cdef:
       np.npy_intp *dims = [n1, max_val+1]
       np.ndarray[np.float64_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
       np.npy_intp *nandims = [n1]
       np.ndarray[np.float_t, ndim=1] nans = PyArray_ZEROS(1, nandims, NPY_float64, 0)
       int ain
       float ai, wt
    if mask is None:
        for iINDEX0 in range(nINDEX0):
            wt = 1 if w is None else w[iINDEX0]
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai != ai:
                    nans[iINDEX1] += wt
                    continue
                if ai < -1e-6:
                    raise ValueError("negative value in bincount")
                ain = int(ai + 0.1)
                if not -1e-6 < ain - ai < 1e-6:
                    raise ValueError("%f is not an integer value" % ai)
                if ain > max_val:
                    raise ValueError("value %i is greater than max_val (%i)" %
                                     (ain, max_val))
                y[iINDEX1, ain] += wt
    elif w is None:
        for iINDEX1 in range(nINDEX1):
            if mask[iINDEX1]:
                for iINDEX0 in range(nINDEX0):
                    ai = a[INDEXALL]
                    if ai != ai:
                        nans[iINDEX1] += 1
                        continue
                    if ai < -1e-6:
                        raise ValueError("negative value in bincount")
                    ain = int(ai + 0.1)
                    if not -1e-6 < ain - ai < 1e-6:
                        raise ValueError("%f is not an integer value" % ai)
                    if ain > max_val:
                        raise ValueError("value %i is greater than max_val (%i)" %
                                         (ain, max_val))
                    y[iINDEX1, ain] += 1
    else:
        for iINDEX1 in range(nINDEX1):
            if mask[iINDEX1]:
                for iINDEX0 in range(nINDEX0):
                    wt = w[iINDEX0]
                    ai = a[INDEXALL]
                    if ai != ai:
                        nans[iINDEX1] += wt
                        continue
                    if ai < -1e-6:
                        raise ValueError("negative value in bincount")
                    ain = int(ai + 0.1)
                    if not -1e-6 < ain - ai < 1e-6:
                        raise ValueError("%f is not an integer value" % ai)
                    if ain > max_val:
                        raise ValueError("value %i is greater than max_val (%i)" %
                                         (ain, max_val))
                    y[iINDEX1, ain] += wt
    return y, nans
"""

sparse = """
@cython.boundscheck(False)
@cython.wraparound(False)
def SPARSE(object a,
           int max_val,
           np.ndarray[np.float_t, ndim=1] w = None,
           np.ndarray[np.int8_t, ndim=1] mask = None):
    '''Bincount that can handle floats (casts to int), handles NaNs and needs to
     be specified a fixed number of values'''

    cdef:
        Py_ssize_t n_rows = a.shape[0]
        Py_ssize_t n_cols = a.shape[1]

    if w is not None and len(w) != n_rows:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.npy_intp *dims = [n_cols, max_val+1]
        np.ndarray[np.float64_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
        np.ndarray[np.float64_t, ndim=1] nans = PyArray_ZEROS(1, dims, NPY_float64, 0)
        float wt

        np.ndarray[np.DTYPE_t, ndim=1] data = a.data
        np.ndarray[int, ndim=1] indices = a.indices
        np.ndarray[int, ndim=1] indptr = a.indptr
        int ri, i, ci
        np.DTYPE_t ai
        int ain
    for ri in range(a.shape[0]):
        wt = 1 if w is None else w[ri]
        for i in range(indptr[ri], indptr[ri + 1]):
            ci = indices[i]
            if mask is None or mask[ci]:
                ai = data[i]
                if ai != ai:
                    nans[indices[i]] += wt
                    continue
                ain = int(ai + 0.1)
                if not -1e-6 < ain - ai < 1e-6:
                    raise ValueError("%f is not an integer value" % ai)
                if ain > max_val:
                    raise ValueError("value %i is greater than max_val (%i)" %
                                     (ain, max_val))
                y[ci, ain] += wt
    return y, nans
"""

floats['loop'] = loop
floats['sparse'] = sparse

# Int dtypes (not axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES

loop = {}
loop[1] = """\
    if w is not None and len(w) != n0:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.npy_intp *dims = [max_val+1]
        np.ndarray[np.float64_t, ndim=1] y = PyArray_ZEROS(1, dims, NPY_float64, 0)
        int ai
    if mask is not None and not mask[0]:
        return y, 0.0
    if w is None:
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai < 0:
                raise ValueError("negative value in bincount")
            if ai > max_val:
                raise ValueError("value %i is greater than max_val (%i)"
                                 % (ai, max_val))
            y[ai] += 1
    else:
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai < 0:
                raise ValueError("negative value in bincount")
            if ai > max_val:
                raise ValueError("value %i is greater than max_val (%i)"
                                 % (ai, max_val))
            y[ai] += w[iINDEX0]
    return y, 0.0
"""

loop[2] = """\
    if w is not None and len(w) != len(a):
        raise ValueError("invalid length of the weight vector")

    cdef:
       np.npy_intp *dims = [n1, max_val+1]
       np.ndarray[np.float_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
       np.npy_intp *nandims = [n1]
       np.ndarray[np.float64_t, ndim=1] nans = PyArray_ZEROS(1, nandims, NPY_float64, 0)
       int ai
       float wt
    if mask is None:
        for iINDEX0 in range(nINDEX0):
            wt = 1 if w is None else w[iINDEX0]
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai < 0:
                    raise ValueError("negative value in bincount")
                if ai > max_val:
                    raise ValueError("value %i is greater than max_val (%i)"
                                     % (ai, max_val))
                y[iINDEX1, ai] += wt
    elif w is None:
        for iINDEX1 in range(nINDEX1):
            if mask[iINDEX1]:
                for iINDEX0 in range(nINDEX0):
                    ai = a[INDEXALL]
                    if ai < 0:
                        raise ValueError("negative value in bincount")
                    if ai > max_val:
                        raise ValueError("value %i is greater than max_val (%i)"
                                         % (ai, max_val))
                    y[iINDEX1, ai] += 1
    else:
        for iINDEX1 in range(nINDEX1):
            if mask[iINDEX1]:
                for iINDEX0 in range(nINDEX0):
                    ai = a[INDEXALL]
                    if ai < 0:
                        raise ValueError("negative value in bincount")
                    if ai > max_val:
                        raise ValueError("value %i is greater than max_val (%i)"
                                         % (ai, max_val))
                    y[iINDEX1, ai] += w[iINDEX0]
    return y, nans
"""

sparse = """
@cython.boundscheck(False)
@cython.wraparound(False)
def SPARSE(object a,
           int max_val,
           np.ndarray[np.float_t, ndim=1] w = None,
           np.ndarray[np.int8_t, ndim=1] mask = None):
    '''Bincount that can handle floats (casts to int), handles NaNs and needs to
     be specified a fixed number of values'''

    cdef:
        Py_ssize_t n_rows = a.shape[0]
        Py_ssize_t n_cols = a.shape[1]

    if w is not None and len(w) != n_rows:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.npy_intp *dims = [n_cols, max_val+1]
        np.ndarray[np.float64_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
        np.ndarray[np.float64_t, ndim=1] nans = PyArray_ZEROS(1, dims, NPY_float64, 0)
        float wt

        np.ndarray[np.DTYPE_t, ndim=1] data = a.data
        np.ndarray[int, ndim=1] indices = a.indices
        np.ndarray[int, ndim=1] indptr = a.indptr
        int ri, i, ci
    for ri in range(a.shape[0]):
        wt = 1 if w is None else w[ri]
        for i in range(indptr[ri], indptr[ri + 1]):
            ci = indices[i]
            if mask is None or mask[ci]:
                y[ci, data[i]] += wt
    return y, nans
"""

ints['loop'] = loop
ints['sparse'] = sparse



# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "bincount"
slow['signature'] = "arr, max_val, weights, mask"
slow['func'] = "bn.slow.bincount(arr, max_val, weights=None, mask=None)"

# Template ------------------------------------------------------------------

bincount = {}
bincount['name'] = 'bincount'
bincount['is_reducing_function'] = False
bincount['cdef_output'] = False
bincount['slow'] = slow
bincount['sparse'] = {}
bincount['templates'] = {}
bincount['templates']['float_None'] = floats
bincount['templates']['int_None'] = ints
bincount['pyx_file'] = 'func/%sbit/bincount.pyx'

bincount['main'] = '''"nanequal auto-generated from template"

def bincount(arr, max_val, weights=None, mask=None):
    """
    Count number of occurrences of each value in array.

    The number of bins is max_val+1. Each bin gives the number of occurrences
    of its index value in `x`. If `weights` is specified the input array is
    weighted by it, i.e. if a value ``n`` is found at position ``i``,
    ``out[n] += weight[i]`` instead of ``out[n] += 1``. A subset of columns
    can be selected by additional argument `mask`.

    Function differs from numpy in that it can handle float arrays; values
    need to be close to integers (allowed error is 1e-6). The function also
    returns a count of NaN values. The maximal value must be given as an
    argument.

    Unlike numpy.bincount, this function also handles 2d arrays.

    Parameters
    ----------
    x : array_like, 1 or 2 dimensions, nonnegative elements
        Input array.
    max_val : int
        The maximal value in the array
    weights : array_like, optional
        Weights, array of the same length as `x`.
    mask: array_like, of type char (interpreted as bool)
        Selects the columns

    Returns
    -------
    out : ndarray of float64, 1- or 2-dimensional
        The result of binning the input array.
    nans: the number of NaNs; a scalar or a 1-d vector of length x.shape[1]

    Raises
    ------
    ValueError
        If the input is not 1- or 2-dimensional, or contains elements that are
        not close enough to integers, negative or grater than max_val, or if the
        length of the weight vector does not match the length of the array

    """
    func, a, weights, mask = bincount_selector(arr, max_val, weights, mask)
    return func(a, max_val, weights, mask)




def bincount_selector(arr, max_val, weights, mask):
    cdef int dtype
    cdef tuple key

    if sp.issparse(arr):
        a = arr
        dtype = PyArray_TYPE(arr.data)
        ndim = 0
        key = (0, dtype, None)
    else:
        if type(arr) is np.ndarray:
            a = arr
        else:
            a = np.array(arr, copy=False)
        dtype = PyArray_TYPE(arr)
        ndim = PyArray_NDIM(a)
        key = (ndim, dtype, None)

    if weights is not None and (
            type(weights) is not np.ndarray or
            weights.dtype is not np.float):
        weights = np.array(weights, copy=False, dtype=np.float)
    if type(mask) is np.ndarray:
        if mask.dtype is not np.int8:
            if mask.dtype.itemsize == 1:
                mask = np.frombuffer(mask, dtype=np.int8)
            else:
                mask = np.array(mask, copy=False, dtype=np.int8)
    elif mask is not None:
        mask = np.array(mask, copy=False, dtype=np.int8)

    try:
        func = bincount_dict[key]
        return func, a, weights, mask
    except KeyError:
        pass

    try:
        func = bincount_slow_dict[None]
    except KeyError:
        tup = (str(ndim), str(a.dtype))
        raise TypeError("Unsupported ndim/dtype (%s/%s)." % tup)
    return func, a, weights, mask
'''



"contingency template"

from copy import deepcopy
import bottlechest as bn

__all__ = ["contingency"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = True
floats['force_output_dtype'] = 'bool'
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a,
                              np.ndarray[np.int8_t, ndim=1] b,
                              int max_val, int max_val2,
                              np.ndarray[np.float_t, ndim=1] w = None,
                              np.ndarray[np.int8_t, ndim=1] mask = None):
    '''Contingency matrix from array. For 2d-data, it returns 3d array
       [variable, value-from-a, value-from-b]. For 3-d, it returns a 2d array'''
"""

loop = {}

loop[1] = """\
    if len(b) != n0:
        raise ValueError("invalid length of the column variable vector")
    if w is not None and len(w) != n0:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.npy_intp *dims = [max_val2 + 1, max_val + 1]
        np.ndarray[np.float64_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
        np.npy_intp *nandims = [max_val2 + 1]
        np.ndarray[np.float_t, ndim=1] nans = PyArray_ZEROS(1, nandims, NPY_float64, 0)
        int ain
        char bin
        float ai, wt
    if mask is not None and not mask[0]:
        return y, nans
    if w is None:
        for iINDEX0 in range(nINDEX0):
            bin = b[iINDEX0]
            if bin > max_val2:
                raise ValueError("value %i is greater than max_val2 (%i)" %
                                 (bin, max_val2))
            ai = a[INDEXALL]
            if ai != ai:
                nans[bin] += 1
                continue
            if ai < -1e-6:
                raise ValueError("negative value in contingency")
            ain = int(ai + 0.1)
            if not -1e-6 < ain - ai < 1e-6:
                raise ValueError("%f is not an integer value" % ai)
            if ain > max_val:
                raise ValueError("value %i is greater than max_val (%i)" %
                                 (ain, max_val))
            y[bin, ain] += 1
    else:
        for iINDEX0 in range(nINDEX0):
            wt = w[iINDEX0]
            ai = a[INDEXALL]
            if ai != ai:
                nans[bin] += wt
                continue
            if ai < -1e-6:
                raise ValueError("negative value in contingency")
            ain = int(ai + 0.1)
            if not -1e-6 < ain - ai < 1e-6:
                raise ValueError("%f is not an integer value" % ai)
            bin = b[iINDEX0]
            if bin > max_val2:
                raise ValueError("value %i is greater than max_val2 (%i)" %
                                 (bin, max_val2))
            y[bin, ain] += wt
    return y, nans
"""

loop[2] = """\
    if len(b) != n0:
        raise ValueError("invalid length of the column variable vector")
    if w is not None and len(w) != len(a):
        raise ValueError("invalid length of the weight vector")

    cdef:
       np.npy_intp *dims = [n1, max_val2 + 1, max_val + 1]
       np.ndarray[np.float64_t, ndim=3] y = PyArray_ZEROS(3, dims, NPY_float64, 0)
       np.npy_intp *nandims = [n1, max_val2 + 1]
       np.ndarray[np.float_t, ndim=2] nans = PyArray_ZEROS(2, nandims, NPY_float64, 0)
       int ain
       char bin
       float ai, wt
    for iINDEX0 in range(nINDEX0):
        wt = 1 if w is None else w[iINDEX0]
        bin = b[iINDEX0]
        if bin > max_val2:
            raise ValueError("value %i is greater than max_val2 (%i)" %
                             (bin, max_val2))
        for iINDEX1 in range(nINDEX1):
            if mask is None or mask[iINDEX1]:
                ai = a[INDEXALL]
                if ai != ai:
                    nans[iINDEX1, bin] += wt
                    continue
                if ai < -1e-6:
                    raise ValueError("negative value in contingency")
                ain = int(ai + 0.1)
                if not -1e-6 < ain - ai < 1e-6:
                    raise ValueError("%f is not an integer value" % ai)
                if ain > max_val:
                    raise ValueError("value %i is greater than max_val (%i)" %
                                     (ain, max_val))
                y[iINDEX1, bin, ain] += wt
    return y, nans
"""

sparse = """
@cython.boundscheck(False)
@cython.wraparound(False)
def SPARSE(object a,
           np.ndarray[np.int8_t, ndim=1] b,
           int max_val, int max_val2,
           np.ndarray[np.float_t, ndim=1] w = None,
           np.ndarray[np.int8_t, ndim=1] mask = None):
    '''contingency that can handle floats (casts to int), handles NaNs and needs to
     be specified a fixed number of values'''

    cdef:
        Py_ssize_t n_rows = a.shape[0]
        Py_ssize_t n_cols = a.shape[1]

    if len(b) != n_rows:
        raise ValueError("invalid length of the column variable vector")
    if w is not None and len(w) != n_rows:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.npy_intp *dims = [n_cols, max_val2 + 1, max_val + 1]
        np.ndarray[np.float64_t, ndim=3] y = PyArray_ZEROS(3, dims, NPY_float64, 0)
        np.npy_intp *nandims = [n_cols, max_val2 + 1]
        np.ndarray[np.float_t, ndim=2] nans = PyArray_ZEROS(2, nandims, NPY_float64, 0)
        float wt

        np.ndarray[np.DTYPE_t, ndim=1] data = a.data
        np.ndarray[int, ndim=1] indices = a.indices
        np.ndarray[int, ndim=1] indptr = a.indptr
        int ri, i, ci
        np.DTYPE_t ai
        int ain
        char bin
    for ri in range(a.shape[0]):
        wt = 1 if w is None else w[ri]
        bin = b[ri]
        if bin > max_val2:
            raise ValueError("value %i is greater than max_val2 (%i)" %
                             (bin, max_val2))
        for i in range(indptr[ri], indptr[ri + 1]):
            ci = indices[i]
            if mask is None or mask[ci]:
                ai = data[i]
                if ai != ai:
                    nans[indices[i], bin] += wt
                    continue
                ain = int(ai + 0.1)
                if not -1e-6 < ain - ai < 1e-6:
                    raise ValueError("%f is not an integer value" % ai)
                if ain > max_val:
                    raise ValueError("value %i is greater than max_val (%i)" %
                                     (ain, max_val))
                y[ci, bin, ain] += wt
    return y, nans
"""

floats['loop'] = loop
floats['sparse'] = sparse

# Int dtypes (not axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES

loop = {}
loop[1] = """\
    if w is not None and len(w) != n0:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.npy_intp *dims = [max_val2 + 1, max_val + 1]
        np.ndarray[np.float64_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
        np.npy_intp *nandims = [max_val2 + 1]
        np.ndarray[np.float_t, ndim=1] nans = PyArray_ZEROS(1, nandims, NPY_float64, 0)
        int ai
        char bin
    if mask is not None and not mask[0]:
        return y, nans
    if w is None:
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai < 0:
                raise ValueError("negative value in contingency")
            if ai > max_val:
                raise ValueError("value %i is greater than max_val (%i)"
                                 % (ai, max_val))
            bin = b[iINDEX0]
            if bin > max_val2:
                raise ValueError("value %i is greater than max_val2 (%i)" %
                                 (bin, max_val2))
            y[bin, ai] += 1
    else:
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai < 0:
                raise ValueError("negative value in contingency")
            if ai > max_val:
                raise ValueError("value %i is greater than max_val (%i)"
                                 % (ai, max_val))
            bin = b[iINDEX0]
            if bin > max_val2:
                raise ValueError("value %i is greater than max_val2 (%i)" %
                                 (bin, max_val2))
            y[bin, ai] += w[iINDEX0]
    return y, nans
"""

loop[2] = """\
    if w is not None and len(w) != len(a):
        raise ValueError("invalid length of the weight vector")

    cdef:
       np.npy_intp *dims = [n1, max_val2 + 1, max_val + 1]
       np.ndarray[np.float_t, ndim=3] y = PyArray_ZEROS(3, dims, NPY_float64, 0)
       np.npy_intp *nandims = [n1, max_val2+1]
       np.ndarray[np.float64_t, ndim=2] nans = PyArray_ZEROS(2, nandims, NPY_float64, 0)
       int ai
       char bin
       float wt
    for iINDEX0 in range(nINDEX0):
        bin = b[iINDEX0]
        if bin > max_val2:
            raise ValueError("value %i is greater than max_val2 (%i)" %
                             (bin, max_val2))
        wt = 1 if w is None else w[iINDEX0]
        for iINDEX1 in range(nINDEX1):
            if mask is None or mask[iINDEX1]:
                ai = a[INDEXALL]
                if ai < 0:
                    raise ValueError("negative value in contingency")
                if ai > max_val:
                    raise ValueError("value %i is greater than max_val (%i)"
                                     % (ai, max_val))
                y[iINDEX1, bin, ai] += wt
    return y, nans
"""

sparse = """
@cython.boundscheck(False)
@cython.wraparound(False)
def SPARSE(object a,
           np.ndarray[np.int8_t, ndim=1] b,
           int max_val, int max_val2,
           np.ndarray[np.float_t, ndim=1] w = None,
           np.ndarray[np.int8_t, ndim=1] mask = None):
    '''contingency that can handle floats (casts to int), handles NaNs and needs to
     be specified a fixed number of values'''

    cdef:
        Py_ssize_t n_rows = a.shape[0]
        Py_ssize_t n_cols = a.shape[1]

    if w is not None and len(w) != n_rows:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.npy_intp *dims = [n_cols, max_val2 + 1, max_val + 1]
        np.ndarray[np.float64_t, ndim=3] y = PyArray_ZEROS(3, dims, NPY_float64, 0)
        np.npy_intp *nandims = [n_cols, max_val2+1]
        np.ndarray[np.float_t, ndim=2] nans = PyArray_ZEROS(2, nandims, NPY_float64, 0)
        float wt
        char bin

        np.ndarray[np.DTYPE_t, ndim=1] data = a.data
        np.ndarray[int, ndim=1] indices = a.indices
        np.ndarray[int, ndim=1] indptr = a.indptr
        int ri, i, ci
    for ri in range(a.shape[0]):
        wt = 1 if w is None else w[ri]
        bin = b[ri]
        if bin > max_val2:
            raise ValueError("value %i is greater than max_val2 (%i)" %
                             (bin, max_val2))
        for i in range(indptr[ri], indptr[ri + 1]):
            ci = indices[i]
            if mask is None or mask[ci]:
                y[ci, bin, data[i]] += wt
    return y, nans
"""

ints['loop'] = loop
ints['sparse'] = sparse



# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "contingency"
slow['signature'] = "arr, b, max_val, max_val2, weights, mask"
slow['func'] = "bn.slow.contingency(arr, b, max_val, max_val2, weights=None, mask=None)"

# Template ------------------------------------------------------------------

contingency = {}
contingency['name'] = 'contingency'
contingency['is_reducing_function'] = False
contingency['cdef_output'] = False
contingency['slow'] = slow
contingency['sparse'] = {}
contingency['templates'] = {}
contingency['templates']['float_None'] = floats
contingency['templates']['int_None'] = ints
contingency['pyx_file'] = 'func/%sbit/contingency.pyx'

contingency['main'] = '''"nanequal auto-generated from template"

def contingency(arr, b, max_val, max_val2, weights=None, mask=None):
    """
    Compute the contingency matrices for each column (excluding the masked)
    versus the vector b.

    If the array is 1-dimensional, a 2d contingency matrix is returned. If the
    array is 2d, the function returns a 3d array, with the first dimension
    corresponding to column index (variable in the input array).

    The rows of contingency matrix correspond to values of variables, the
    columns correspond to values in vector `b`. The number of rows is
    `max_val+1`, the nuber of columns is `max_val2+1`.

    Rows in the input array can be weights (argument `weights`). A subset of
    columns can be selected by additional argument `mask`.

    The function also returns a count of NaN values per each value of `b`.

    Parameters
    ----------
    x : array_like, 1 or 2 dimensions, nonnegative elements
        Input array.
    b : array_like, 1 dimension, of type int8
        The values of the second variable.
    max_val : int
        The maximal value in the array
    max_val2 : int
        The maximal value in `b`
    weights : array_like, optional
        Weights, array of the same length as `x`.
    mask: array_like, of type char (interpreted as bool)
        Selects the columns

    Returns
    -------
    out : ndarray of ints, 1- or 2-dimensional
        The result of binning the input array.
    nans: the number of NaNs; a 1-d vector of length `max_val`, or 2-d array `(max_val2, x.shape[1])`

    Raises
    ------
    ValueError
        If the input is not 1- or 2-dimensional, or contains elements that are
        not close enough to integers, negative or grater than max_val, or if the
        length of the weight vector does not match the length of the array

    """
    func, a, b, weights, mask = contingency_selector(arr, b, max_val, max_val2, weights, mask)
    return func(a, b, max_val, max_val2, weights, mask)




def contingency_selector(arr, b, max_val, max_val2, weights, mask):
    cdef int dtype
    cdef tuple key

    if sp.issparse(arr):
        a = arr
        dtype = PyArray_TYPE(arr.data)
        ndim = 0
        key = (0, dtype, None)
    else:
        if type(arr) is np.ndarray:
            a = arr
        else:
            a = np.array(arr, copy=False)
        dtype = PyArray_TYPE(arr)
        ndim = PyArray_NDIM(a)
        key = (ndim, dtype, None)

    if type(b) is np.ndarray:
        if b.dtype != np.int8:
            b = b.astype(np.int8)
    else:
        b = np.array(b, copy=False, dtype=np.int8)
    if weights is not None and (
            type(weights) is not np.ndarray or
            weights.dtype is not np.float):
        weights = np.array(weights, copy=False, dtype=np.float)
    if type(mask) is np.ndarray:
        if mask.dtype is not np.int8:
            if mask.dtype.itemsize == 1:
                mask = np.frombuffer(mask, dtype=np.int8)
            else:
                mask = np.array(mask, copy=False, dtype=np.int8)
    elif mask is not None:
        mask = np.array(mask, copy=False, dtype=np.int8)

    try:
        func = contingency_dict[key]
        return func, a, b, weights, mask
    except KeyError:
        pass

    try:
        func = contingency_slow_dict[None]
    except KeyError:
        tup = (str(ndim), str(a.dtype))
        raise TypeError("Unsupported ndim/dtype (%s/%s)." % tup)
    return func, a, b, weights, mask
'''



"Test moving window functions."

# For support of python 2.5
from __future__ import with_statement

from nose.tools import assert_true
import numpy as np
from numpy.testing import (assert_equal, assert_array_equal,
                           assert_array_almost_equal)
nan = np.nan
import bottlechest as bn


def arrays(dtypes=bn.dtypes, nans=True):
    "Iterator that yield arrays to use for unit testing."
    try:
        import scipy
        SCIPY = True
    except ImportError:
        SCIPY = False
    ss = {}
    ss[1] = {'size':  4, 'shapes': [(4,)]}
    ss[2] = {'size':  6, 'shapes': [(1,6), (2,3)]}
    ss[3] = {'size':  6, 'shapes': [(1,2,3)]}
    if SCIPY:
        # Unaccelerated fallback requires scipy
        ss[4] = {'size': 24, 'shapes': [(1,2,3,4)]}  # Unaccelerated
    for ndim in ss:
        size = ss[ndim]['size']
        shapes = ss[ndim]['shapes']
        for dtype in dtypes:
            a = np.arange(size, dtype=dtype)
            for shape in shapes:
                a = a.reshape(shape)
                yield a
                yield -a
            if issubclass(a.dtype.type, np.inexact): 
                if nans:
                    for i in range(a.size):
                        a.flat[i] = np.nan
                        yield a
                        yield -a

def unit_maker(func, func0, decimal=np.inf, nans=True):
    "Test that bn.xxx gives the same output as a reference function."
    msg = '\nfunc %s | window %d | input %s (%s) | shape %s | axis %s\n'
    msg += '\nInput array:\n%s\n'
    for i, arr in enumerate(arrays(nans=nans)):
        for axis in range(-arr.ndim, arr.ndim):
            windows = range(1, arr.shape[axis])
            if len(windows) == 0:
                windows = [1]
            for window in windows:
                with np.errstate(invalid='ignore'):
                    actual = func(arr, window, axis=axis)
                    desired = func0(arr, window, axis=axis, method='loop')
                tup = (func.__name__, window, 'a'+str(i), str(arr.dtype),
                       str(arr.shape), str(axis), arr)
                err_msg = msg % tup
                if (decimal < np.inf) and (np.isfinite(arr).sum() > 0):
                    assert_array_almost_equal(actual, desired, decimal,
                                              err_msg)
                else:
                    assert_array_equal(actual, desired, err_msg)
                err_msg += '\n dtype mismatch %s %s'
                if hasattr(actual, 'dtype') or hasattr(desired, 'dtype'):
                    da = actual.dtype
                    dd = desired.dtype
                    assert_equal(da, dd, err_msg % (da, dd))

def test_move_sum():
    "Test move_sum."
    yield unit_maker, bn.move_sum, bn.slow.move_sum, 5

def test_move_nansum():
    "Test move_nansum."
    yield unit_maker, bn.move_nansum, bn.slow.move_nansum, 5

def test_move_mean():
    "Test move_mean."
    yield unit_maker, bn.move_mean, bn.slow.move_mean, 5

def test_move_median():
    "Test move_median."
    yield unit_maker, bn.move_median, bn.slow.move_median, 5, False

def test_move_nanmean():
    "Test move_nanmean."
    yield unit_maker, bn.move_nanmean, bn.slow.move_nanmean, 5

def test_move_std():
    "Test move_std."
    yield unit_maker, bn.move_std, bn.slow.move_std, 5

def test_move_nanstd():
    "Test move_nanstd."
    yield unit_maker, bn.move_nanstd, bn.slow.move_nanstd, 5

def test_move_min():
    "Test move_min."
    yield unit_maker, bn.move_min, bn.slow.move_min, 5

def test_move_max():
    "Test move_max."
    yield unit_maker, bn.move_max, bn.slow.move_max, 5

def test_move_nanmin():
    "Test move_nanmin."
    yield unit_maker, bn.move_nanmin, bn.slow.move_nanmin, 5

def test_move_nanmax():
    "Test move_nanmax."
    yield unit_maker, bn.move_nanmax, bn.slow.move_nanmax, 5

# ----------------------------------------------------------------------------
# Regression test for square roots of negative numbers

def test_move_std_sqrt():
    "Test move_std for neg sqrt."

    a = [0.0011448196318903589,
         0.00028718669878572767,
         0.00028718669878572767,
         0.00028718669878572767,
         0.00028718669878572767]
    err_msg = "Square root of negative number. ndim = %d"
    b = bn.move_std(a, window=3)
    assert_true(np.isfinite(b[2:]).all(), err_msg % 1)

    a2 = np.array([a, a])
    b = bn.move_std(a2, window=3, axis=1)
    assert_true(np.isfinite(b[:, 2:]).all(), err_msg % 2)

    a3 = np.array([[a, a], [a, a]])
    b = bn.move_std(a3, window=3, axis=2)
    assert_true(np.isfinite(b[:, :, 2:]).all(), err_msg % 3)

def test_move_nanstd_sqrt():
    "Test move_nanstd for neg sqrt."

    a = [0.0011448196318903589,
         0.00028718669878572767,
         0.00028718669878572767,
         0.00028718669878572767,
         0.00028718669878572767]
    err_msg = "Square root of negative number. ndim = %d"
    b = bn.move_nanstd(a, window=3)
    assert_true(np.isfinite(b[2:]).all(), err_msg % 1)

    a2 = np.array([a, a])
    b = bn.move_nanstd(a2, window=3, axis=1)
    assert_true(np.isfinite(b[:, 2:]).all(), err_msg % 2)

    a3 = np.array([[a, a], [a, a]])
    b = bn.move_nanstd(a3, window=3, axis=2)
    assert_true(np.isfinite(b[:, :, 2:]).all(), err_msg % 3)



"stats template"

# For computation of weighted variance, see
# http://en.wikipedia.org/wiki/Weighted_sample_variance#Weighted_sample_variance

from copy import deepcopy
import bottlechest as bn

__all__ = ["stats"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = True
floats['force_output_dtype'] = 'bool'
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a,
                              np.ndarray[np.float_t, ndim=1] w = None,
                              int compute_variance = False):
    '''Compute min, max, mean and #nans.
    '''
"""

loop = {}

loop[1] = """\
    cdef:
        np.DTYPE_t ai
        np.float_t wt
        np.DTYPE_t a_min = MAXDTYPE
        np.DTYPE_t a_max = MINDTYPE
        np.float_t mean = 0
        np.float_t var = 0
        np.float_t non_nans = 0
        np.float_t nans = 0
        np.float_t tot_wt2, d
    if w is None:
        for iINDEX0 in range(nINDEX0):
            ai = a[iINDEX0]
            if ai != ai:
                nans += 1
                continue
            if ai < a_min:
                a_min = ai
            if ai > a_max:
                a_max = ai
            mean += ai
        non_nans = nINDEX0 - nans
        if non_nans != 0:
            mean /= non_nans
        if compute_variance and non_nans >= 2:
            for iINDEX0 in range(nINDEX0):
                ai = a[iINDEX0]
                if ai == ai:
                    var += (ai - mean) ** 2
            var /= non_nans - 1
    else:
        if len(w) != n0:
            raise ValueError("invalid length of the weight vector ({} != {})".
                format(len(w), n0))
        for iINDEX0 in range(nINDEX0):
            ai = a[iINDEX0]
            wt = w[iINDEX0]
            if ai != ai:
                nans += wt
                continue
            else:
                non_nans += wt
            if ai < a_min:
                a_min = ai
            if ai > a_max:
                a_max = ai
            mean += wt * ai
        if non_nans != 0:
            mean /= non_nans
            if compute_variance:
                tot_wt2 = 0
                for iINDEX0 in range(nINDEX0):
                    ai = a[iINDEX0]
                    if ai == ai:
                        wt = w[iINDEX0]
                        tot_wt2 += wt ** 2
                        var += wt * (ai - mean) ** 2
                d = non_nans ** 2 - tot_wt2
                if d > 1e-6:
                    var *= non_nans / d
    return a_min, a_max, mean, var, nans, non_nans
"""

loop[2] = """\
    cdef:
       np.npy_intp *dims = [nINDEX1, 6]
       np.ndarray[np.float64_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
       np.DTYPE_t ai
       np.float64_t mean
       np.float_t wt
       np.float_t var
       np.float_t tot_wt
       np.ndarray[np.float64_t, ndim=1] tot_wt2 = PyArray_ZEROS(1, dims, NPY_float64, 0)
       np.float_t d

    for iINDEX1 in range(nINDEX1):
       y[iINDEX1, 0] = MAXfloat64
       y[iINDEX1, 1] = MINfloat64
    if w is None:
        for iINDEX0 in range(nINDEX0):
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai != ai:
                    y[iINDEX1, 4] += 1
                    continue
                if ai < y[iINDEX1, 0]:
                    y[iINDEX1, 0] = ai
                if ai > y[iINDEX1, 1]:
                    y[iINDEX1, 1] = ai
                y[iINDEX1, 2] += ai
        for iINDEX1 in range(nINDEX1):
            y[iINDEX1, 5] = nINDEX0 - y[iINDEX1, 4]
            if y[iINDEX1, 5] > 0:
                y[iINDEX1, 2] /= y[iINDEX1, 5]
        if compute_variance:
            for iINDEX0 in range(nINDEX0):
                for iINDEX1 in range(nINDEX1):
                    ai = a[INDEXALL]
                    if ai == ai:
                        y[iINDEX1, 3] += (ai - y[iINDEX1, 2]) ** 2
            for iINDEX1 in range(nINDEX1):
                if y[iINDEX1, 5] >= 2:
                    y[iINDEX1, 3] /= y[iINDEX1, 5] - 1
    else:
        for iINDEX0 in range(nINDEX0):
            wt = w[iINDEX0]
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai != ai:
                    y[iINDEX1, 4] += wt
                    continue
                y[iINDEX1, 5] += wt
                if ai < y[iINDEX1, 0]:
                    y[iINDEX1, 0] = ai
                if ai > y[iINDEX1, 1]:
                    y[iINDEX1, 1] = ai
                y[iINDEX1, 2] += wt * ai
        for iINDEX1 in range(nINDEX1):
            if y[iINDEX1, 5] > 0:
                y[iINDEX1, 2] /= y[iINDEX1, 5]
        if compute_variance:
            for iINDEX0 in range(nINDEX0):
                wt = w[iINDEX0]
                for iINDEX1 in range(nINDEX1):
                    if y[iINDEX1, 5] >= 2:
                        ai = a[INDEXALL]
                        if ai == ai:
                            tot_wt2[iINDEX1] += wt ** 2
                            y[iINDEX1, 3] += wt * (ai - y[iINDEX1, 2]) ** 2
            for iINDEX1 in range(nINDEX1):
                tot_wt = y[iINDEX1, 5]
                d = tot_wt ** 2 - tot_wt2[iINDEX1]
                if d > 1e-6:
                    y[iINDEX1, 3] *= tot_wt / d
    return y
"""

sparse = """
@cython.boundscheck(False)
@cython.wraparound(False)
def SPARSE(object a,
           np.ndarray[np.float_t, ndim=1] w = None,
           int compute_variance = False):
    '''Compute min, max, #nans, mean and variance.
    '''

    cdef:
        Py_ssize_t n_rows = a.shape[0]
        Py_ssize_t n_cols = a.shape[1]

    if w is not None and len(w) != n_rows:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.ndarray[np.DTYPE_t, ndim=1] data = a.data
        np.ndarray[int, ndim=1] indices = a.indices
        np.ndarray[int, ndim=1] indptr = a.indptr

        np.npy_intp *dims = [n_cols, 6]
        np.ndarray[np.float64_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
        np.ndarray[np.float64_t, ndim=1] tot_wt2 = PyArray_ZEROS(1, dims, NPY_float64, 0)

        int ri, ci
        np.float_t wt
        np.float_t tot_w = 0
        np.float_t d
        np.DTYPE_t ai

    for ci in range(n_cols):
       y[ci, 0] = MAXfloat64
       y[ci, 1] = MINfloat64
    if w is None:
       tot_w = n_rows
    else:
        for ri in range(n_rows):
            tot_w += w[ri]
    if tot_w == 0:
        return y

    for ri in range(a.shape[0]):
        wt = 1 if w is None else w[ri]
        for i in range(indptr[ri], indptr[ri + 1]):
            ai = data[i]
            if ai != ai:
                continue
            ci = indices[i]
            y[ci, 5] += wt
            if ai < y[ci, 0]:
                y[ci, 0] = ai
            if ai > y[ci, 1]:
                y[ci, 1] = ai
            y[ci, 2] += wt * ai
    for ci in range(n_cols):
        y[ci, 4] = tot_w - y[ci, 5]
        if y[ci, 5] != 0:
            y[ci, 2] /= y[ci, 5]
    if compute_variance:
        for ri in range(a.shape[0]):
            wt = 1 if w is None else w[ri]
            for i in range(indptr[ri], indptr[ri + 1]):
                ai = data[i]
                if ai == ai:
                    ci = indices[i]
                    y[ci, 3] += wt * (ai - y[ci, 2]) ** 2
                    tot_wt2[ci] += wt ** 2
        for ci in range(n_cols):
            d = y[ci, 5] ** 2 - tot_wt2[ci]
            if d > 1e-6:
                y[ci, 3] *= y[ci, 5] / d
    return y
"""

floats['loop'] = loop
floats['sparse'] = sparse

# Int dtypes (not axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES

loop = {}
loop[1] = """\
    cdef:
        np.DTYPE_t ai
        np.float_t wt
        np.DTYPE_t a_min = MAXDTYPE
        np.DTYPE_t a_max = MINDTYPE
        np.float64_t mean = 0
        np.float64_t var = 0
        np.float_t tot_w, tot_w2, d
    if n0 == 0:
        return (a_min, a_max, 0, 0, 0, 0)
    if w is None:
        tot_w = nINDEX0
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            if ai < a_min:
                a_min = ai
            if ai > a_max:
                a_max = ai
            mean += ai
        mean /= n0
        if compute_variance and n0 >= 2:
            for iINDEX0 in range(nINDEX0):
                var += (a[INDEXALL] - mean) ** 2
            var /= n0 - 1
    else:
        tot_w = 0
        if len(w) != n0:
            raise ValueError("invalid length of the weight vector")
        for iINDEX0 in range(nINDEX0):
            ai = a[INDEXALL]
            wt = w[iINDEX0]
            tot_w += wt
            if ai < a_min:
                a_min = ai
            if ai > a_max:
                a_max = ai
            mean += wt * ai
        if tot_w != 0:
            mean /= tot_w
        if compute_variance:
            tot_w2 = 0
            for iINDEX0 in range(nINDEX0):
                tot_w2 += w[iINDEX0] ** 2
                var += w[iINDEX0] * (a[INDEXALL] - mean) ** 2
            d = tot_w ** 2 - tot_w2
            if d > 1e-6:
                var *= tot_w / d
    return a_min, a_max, mean, var, 0, tot_w
"""

loop[2] = """\
    cdef:
       np.npy_intp *dims = [n1, 6]
       np.ndarray[np.float64_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
       np.DTYPE_t ai
       np.float64_t mean
       np.float_t wt
       np.float_t tot_w = 0, tot_w2 = 0

    if w is None:
       tot_w = nINDEX0
    else:
        for iINDEX0 in range(nINDEX0):
            tot_w += w[iINDEX0]

    for iINDEX1 in range(nINDEX1):
       y[iINDEX1, 0] = MAXfloat64
       y[iINDEX1, 1] = MINfloat64
       y[iINDEX1, 5] = tot_w

    if tot_w == 0:
        return y

    if w is None:
        for iINDEX0 in range(nINDEX0):
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai < y[iINDEX1, 0]:
                    y[iINDEX1, 0] = ai
                if ai > y[iINDEX1, 1]:
                    y[iINDEX1, 1] = ai
                y[iINDEX1, 2] += ai
        for iINDEX1 in range(nINDEX1):
            y[iINDEX1, 2] /= nINDEX0
            mean = y[iINDEX1, 2]
        if compute_variance and nINDEX0 >= 2:
            for iINDEX0 in range(nINDEX0):
                for iINDEX1 in range(nINDEX1):
                    y[iINDEX1, 3] += (a[INDEXALL] - y[iINDEX1, 2]) ** 2
        for iINDEX1 in range(nINDEX1):
            y[iINDEX1, 3] /= nINDEX0 - 1
    else:
        for iINDEX0 in range(nINDEX0):
            wt = w[iINDEX0]
            for iINDEX1 in range(nINDEX1):
                ai = a[INDEXALL]
                if ai < y[iINDEX1, 0]:
                    y[iINDEX1, 0] = ai
                if ai > y[iINDEX1, 1]:
                    y[iINDEX1, 1] = ai
                y[iINDEX1, 2] += wt * ai
        for iINDEX1 in range(nINDEX1):
            y[iINDEX1, 2] /= tot_w
            mean = y[iINDEX1, 2]
        if compute_variance:
            for iINDEX0 in range(nINDEX0):
                wt = w[iINDEX0]
                tot_w2 += wt ** 2
                for iINDEX1 in range(nINDEX1):
                    y[iINDEX1, 3] += wt * (a[INDEXALL] - y[iINDEX1, 2]) ** 2
            d = tot_w ** 2 - tot_w2
            if d > 1e-6:
                for iINDEX1 in range(nINDEX1):
                    y[iINDEX1, 3] *= tot_w / d
    return y
"""

sparse = """
@cython.boundscheck(False)
@cython.wraparound(False)
def SPARSE(object a,
           np.ndarray[np.float_t, ndim=1] w = None,
           int compute_variance = False):
    '''Compute min, max, #nans, mean and variance.
    '''

    cdef:
        Py_ssize_t n_rows = a.shape[0]
        Py_ssize_t n_cols = a.shape[1]

    if w is not None and len(w) != n_rows:
        raise ValueError("invalid length of the weight vector")

    cdef:
        np.ndarray[np.DTYPE_t, ndim=1] data = a.data
        np.ndarray[int, ndim=1] indices = a.indices
        np.ndarray[int, ndim=1] indptr = a.indptr

        np.npy_intp *dims = [n_cols, 6]
        np.ndarray[np.float64_t, ndim=2] y = PyArray_ZEROS(2, dims, NPY_float64, 0)
        np.ndarray[np.float64_t, ndim=1] tot_w2 = PyArray_ZEROS(1, dims, NPY_float64, 0)
        np.float_t wt
        np.float_t tot_w = 0
        int ri, ci
        np.DTYPE_t ai

    for ci in range(n_cols):
       y[ci, 0] = MAXfloat64
       y[ci, 1] = MINfloat64

    if w is None:
       tot_w = n_rows
    else:
        for ri in range(n_rows):
            tot_w += w[ri]
    if tot_w == 0:
        return y

    for ri in range(a.shape[0]):
        wt = 1 if w is None else w[ri]
        for i in range(indptr[ri], indptr[ri + 1]):
            ci = indices[i]
            ai = data[i]
            if ai < y[ci, 0]:
                y[ci, 0] = ai
            if ai > y[ci, 1]:
                y[ci, 1] = ai
            y[ci, 5] += wt
            y[ci, 2] += wt * ai
    for ci in range(n_cols):
        y[ci, 4] = tot_w - y[ci, 5]
        if y[ci, 5] != 0:
            y[ci, 2] /= y[ci, 5]
    if compute_variance:
        for ri in range(a.shape[0]):
            wt = 1 if w is None else w[ri]
            for i in range(indptr[ri], indptr[ri + 1]):
                ci = indices[i]
                tot_w2[ci] += wt ** 2
                y[ci, 3] += wt * (data[i] - y[ci, 2]) ** 2
        for ci in range(n_cols):
            d = y[ci, 5] ** 2 - tot_w2[ci]
            if d > 1e-6:
                y[ci, 3] *= y[ci, 5] / d
    return y
"""

ints['loop'] = loop
ints['sparse'] = sparse



# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "stats"
slow['signature'] = "arr, weights, compute_variance"
slow['func'] = "bn.slow.stats(arr, weights=None, compute_variance=False)"

# Template ------------------------------------------------------------------

stats = {}
stats['name'] = 'stats'
stats['is_reducing_function'] = False
stats['cdef_output'] = False
stats['slow'] = slow
stats['sparse'] = {}
stats['templates'] = {}
stats['templates']['float_None'] = floats
stats['templates']['int_None'] = ints
stats['pyx_file'] = 'func/%sbit/stats.pyx'

stats['main'] = '''"stats auto-generated from template"

def stats(arr, weights=None, compute_variance=False):
    """
    Compute min, max, #nans, mean and variance.

    Result is a tuple (min, max, mean, variance, #nans, #non-nans) or an
    array of shape (len(arr), 6).

    The mean and the number of nans and non-nans are weighted.

    Computation of variance requires an additional pass and is not enabled
    by default. Zeros are filled in instead of variance.

    Parameters
    ----------
    x : array_like, 1 or 2 dimensions
        Input array.
    weights : array_like, optional
        Weights, array of the same length as `x`.
    compute_variance : bool, optional
        If set to True, the function also computes variance.

    Returns
    -------
    out : a 6-element tuple or an array of shape (len(x), 6)
        Computed (min, max, mean, 0, #nans and #non-nans)

    Raises
    ------
    ValueError
        If the length of the weight vector does not match the length of the
        array
    """
    func, a, weights = stats_selector(arr, weights)
    return func(a, weights, compute_variance)




def stats_selector(arr, weights):
    cdef int dtype
    cdef tuple key

    if sp.issparse(arr):
        a = arr
        dtype = PyArray_TYPE(arr.data)
        ndim = 0
        key = (0, dtype, None)
    else:
        if type(arr) is np.ndarray:
            a = arr
        else:
            a = np.array(arr, copy=False)
        dtype = PyArray_TYPE(arr)
        ndim = PyArray_NDIM(a)
        key = (ndim, dtype, None)

    if weights is not None and (
            type(weights) is not np.ndarray or
            weights.dtype is not np.float):
        weights = np.array(weights, copy=False, dtype=np.float)

    try:
        func = stats_dict[key]
        return func, a, weights
    except KeyError:
        pass

    try:
        func = stats_slow_dict[None]
    except KeyError:
        tup = (str(ndim), str(a.dtype))
        raise TypeError("Unsupported ndim/dtype (%s/%s)." % tup)
    return func, a, weights
'''



#!/usr/bin/env python

import os
from setuptools import setup
from setuptools.extension import Extension
import numpy as np


CLASSIFIERS = ["Development Status :: 4 - Beta",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: BSD License",
               "Operating System :: OS Independent",
               "Programming Language :: Cython",
               "Programming Language :: Python",
               "Programming Language :: Python :: 3",
               "Topic :: Scientific/Engineering"]

# Description
description = "Fast NumPy array functions written in Cython"
fid = open('README.rst', 'r')
long_description = fid.read()
fid.close()
idx = max(0, long_description.find("Bottlechest is a collection"))
long_description = long_description[idx:]

# Get bottlechest version
ver_file = os.path.join('bottlechest', 'version.py')
fid = open(ver_file, 'r')
VER = fid.read()
fid.close()
VER = VER.split("= ")
VER = VER[1].strip()
VER = VER.strip("\"")
VER = VER.split('.')

NAME                = 'Bottlechest'
MAINTAINER          = "Bioinformatics Laboratory, FRI UL"
MAINTAINER_EMAIL    = "contact@orange.biolab.si"
DESCRIPTION         = description
LONG_DESCRIPTION    = long_description
URL                 = "https://github.com/biolab/bottlechest"
DOWNLOAD_URL        = "https://github.com/biolab/bottlechest"
LICENSE             = "Simplified BSD"
CLASSIFIERS         = CLASSIFIERS
AUTHOR              = "Bioinformatics Laboratory, FRI UL"
AUTHOR_EMAIL        = "contact@orange.biolab.si"
PLATFORMS           = "OS Independent"
MAJOR               = VER[0]
MINOR               = VER[1]
MICRO               = VER[2]
ISRELEASED          = False
VERSION             = '%s.%s.%s' % (MAJOR, MINOR, MICRO)
PACKAGES            = ["bottlechest",
                       "bottlechest/slow",
                       "bottlechest/tests",
                       "bottlechest/benchmark",
                       "bottlechest/src",
                       "bottlechest/src/func",
                       "bottlechest/src/template",
                       "bottlechest/src/template/func"
                      ]
PACKAGE_DATA        = {'bottlechest': ['LICENSE']}
REQUIRES            = ["numpy"]


# Is the default numpy int 32 or 64 bits?
if np.int_ == np.int32:
    bits = '32'
elif np.int_ == np.int64:
    bits = '64'
else:
    raise ValueError("Your OS does not appear to be 32 or 64 bits.")

cfiles = [ a[:-2] for a in  os.listdir("bottlechest/src/func/%sbit/" % bits) \
    if a.endswith(".c") ]

extensions = [ Extension(cf,
               sources=["bottlechest/src/func/%sbit/%s.c" % (bits, cf)],
               include_dirs=[np.get_include()]) for cf in cfiles ]

setup(name=NAME,
      maintainer=MAINTAINER,
      maintainer_email=MAINTAINER_EMAIL,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      url=URL,
      download_url=DOWNLOAD_URL,
      license=LICENSE,
      classifiers=CLASSIFIERS,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      platforms=PLATFORMS,
      version=VERSION,
      packages=PACKAGES,
      package_data=PACKAGE_DATA,
      requires=REQUIRES,
      ext_package='bottlechest',
      ext_modules=extensions
      )



from __future__ import absolute_import

from bottlechest.src.template.template import template
import bottlechest.src.template.template as tempmod
import os.path

from collections import defaultdict
import importlib

HEADER = """#cython: embedsignature=True

import numpy as np
cimport numpy as np
import scipy.sparse as sp
import cython
from numpy cimport NPY_INT8 as NPY_int8
from numpy cimport NPY_INT32 as NPY_int32
from numpy cimport NPY_INT64 as NPY_int64
from numpy cimport NPY_FLOAT32 as NPY_float32
from numpy cimport NPY_FLOAT64 as NPY_float64
from numpy cimport (PyArray_EMPTY, PyArray_TYPE, PyArray_NDIM,
                    PyArray_SIZE, PyArray_DIMS, import_array,
                    PyArray_ArgSort, NPY_QUICKSORT, NPY_CORDER,
                    PyArray_Ravel, PyArray_FillWithScalar, PyArray_Copy,
                    NPY_BOOL,
                    PyArray_ZEROS)

# NPY_INTP is missing from numpy.pxd in cython 0.14.1 and earlier
cdef extern from "numpy/arrayobject.h":
    cdef enum NPY_TYPES:
        NPY_intp "NPY_INTP"

import_array()
import bottlechest as bn

cdef double NAN = <double> np.nan

cdef np.int8_t MINint8 = np.iinfo(np.int8).min
cdef np.int32_t MINint32 = np.iinfo(np.int32).min
cdef np.int64_t MINint64 = np.iinfo(np.int64).min
cdef np.float32_t MINfloat32 = np.NINF
cdef np.float64_t MINfloat64 = np.NINF

cdef np.int8_t MAXint8 = np.iinfo(np.int8).max
cdef np.int32_t MAXint32 = np.iinfo(np.int32).max
cdef np.int64_t MAXint64 = np.iinfo(np.int64).max
cdef np.float32_t MAXfloat32 = np.inf
cdef np.float64_t MAXfloat64 = np.inf

int8 = np.dtype(np.int8)
int32 = np.dtype(np.int32)
int64 = np.dtype(np.int64)
float32 = np.dtype(np.float32)
float64 = np.dtype(np.float64)

if np.int_ == np.int32:
    NPY_int_ = NPY_int32
elif np.int_ == np.int64:
    NPY_int_ = NPY_int64
else:
    raise RuntimeError('Expecting default NumPy int to be 32 or 64 bit.')

cdef extern from "math.h":
    double sqrt(double x)

PARTSORT_ERR_MSG = "`n` (=%d) must be between 1 and %d, inclusive."

"""

def single(templatepy, bits):
    b = importlib.import_module("bottlechest.src.template.%s" % (templatepy))
    funcs = []
    for f in b.__all__:
        funcs.append(b.__dict__[f])
    template(funcs, bits, HEADER)

def package(package, bits=None):
 
    pyfolder = os.path.join(os.path.dirname(tempmod.__file__), package)

    pyfiles = set([ a[:-3] for a in os.listdir(pyfolder) if a.endswith(".py") \
        and a != "func.py" and a != "__init__.py" ])

    for a in sorted(list(pyfiles)):
        try:
            single(package + "." + a, bits)
        except ImportError:
            print("Could not import", a + ". Skipping.")



"nanargmin template"

from copy import deepcopy
import bottlechest as bn

__all__ = ["nanargmin"]

FLOAT_DTYPES = [x for x in bn.dtypes if 'float' in x]
INT_DTYPES = [x for x in bn.dtypes if 'int' in x]

# Float dtypes (not axis=None) ----------------------------------------------

floats = {}
floats['dtypes'] = FLOAT_DTYPES
floats['axisNone'] = False
floats['force_output_dtype'] = 'intp'
floats['reuse_non_nan_func'] = False

floats['top'] = """
@cython.boundscheck(False)
@cython.wraparound(False)
def NAME_NDIMd_DTYPE_axisAXIS(np.ndarray[np.DTYPE_t, ndim=NDIM] a):
    "Index of max of NDIMd, DTYPE array along axis=AXIS ignoring NaNs."
    cdef int allnan = 1
    cdef np.DTYPE_t amin, ai
    cdef Py_ssize_t idx = 0
"""

loop = {}
loop[1] = """\
    if nINDEX0 == 0:
        msg = "numpy.nanargmin raises on a.shape[axis]==0; Bottleneck too."
        raise ValueError(msg)
    amin = MAXDTYPE
    for iINDEX0 in range(nINDEX0 - 1, -1, -1):
        ai = a[INDEXALL]
        if ai <= amin:
            amin = ai
            allnan = 0
            idx = iINDEX0
    if allnan == 0:       
        return np.intp(idx)
    else:
        return NAN
"""
loop[2] = """\
    if nINDEX1 == 0:
        msg = "numpy.nanargmin raises on a.shape[axis]==0; Bottleneck too."
        raise ValueError(msg)
    for iINDEX0 in range(nINDEX0 - 1, -1, -1):
        amin = MAXDTYPE
        allnan = 1
        for iINDEX1 in range(nINDEX1 - 1, -1, -1):
            ai = a[INDEXALL]
            if ai <= amin:
                amin = ai
                allnan = 0
                idx = iINDEX1
        if allnan == 0:       
            y[INDEXPOP] = idx
        else:
            raise ValueError(CANNOTCONVERT)
    return y
"""

floats['loop'] = loop

# Int dtypes (not axis=None) ------------------------------------------------

ints = deepcopy(floats)
ints['dtypes'] = INT_DTYPES 

loop = {}
loop[1] = """\
    if nINDEX0 == 0:
        msg = "numpy.nanargmin raises on a.shape[axis]==0; Bottleneck too."
        raise ValueError(msg)
    amin = MAXDTYPE
    for iINDEX0 in range(nINDEX0):
        ai = a[INDEXALL]
        if ai <= amin:
            amin = ai
            idx = iINDEX0
    return np.intp(idx)
"""
loop[2] = """\
    if nINDEX1 == 0:
        msg = "numpy.nanargmin raises on a.shape[axis]==0; Bottleneck too."
        raise ValueError(msg)
    for iINDEX0 in range(nINDEX0 - 1, -1, -1):
        amin = MAXDTYPE
        for iINDEX1 in range(nINDEX1 - 1, -1, -1):
            ai = a[INDEXALL]
            if ai <= amin:
                amin = ai
                idx = iINDEX1
        y[INDEXPOP] = idx
    return y
"""

ints['loop'] = loop

# Slow, unaccelerated ndim/dtype --------------------------------------------

slow = {}
slow['name'] = "nanargmin"
slow['signature'] = "arr"
slow['func'] = "bn.slow.nanargmin(arr, axis=AXIS)"

# Template ------------------------------------------------------------------

nanargmin = {}
nanargmin['name'] = 'nanargmin'
nanargmin['is_reducing_function'] = True
nanargmin['cdef_output'] = True
nanargmin['slow'] = slow
nanargmin['templates'] = {}
nanargmin['templates']['float'] = floats
nanargmin['templates']['int'] = ints
nanargmin['pyx_file'] = 'func/%sbit/nanargmin.pyx'

nanargmin['main'] = '''"nanargmin auto-generated from template"

CANNOTCONVERT = "Bottleneck copies NumPy bahavior: "
CANNOTCONVERT += "'cannot convert float NaN to integer'"

def nanargmin(arr, axis=None):
    """
    Indices of the minimum values along an axis, ignoring NaNs.
    
    Parameters
    ----------
    a : array_like
        Input data.
    axis : {int, None}, optional
        Axis along which to operate. By default (axis=None) flattened input
        is used.
   
    See also
    --------
    bottlechest.nanargmax: Indices of the maximum values along an axis.
    bottlechest.nanmin: Minimum values along specified axis, ignoring NaNs.

    Returns
    -------
    index_array : ndarray
        An array of indices or a single index value.
    
    Examples
    --------
    >>> a = np.array([[np.nan, 4], [2, 3]])
    >>> bn.nanargmin(a)
    2
    >>> a.flat[1]
    2.0
    >>> bn.nanargmax(a, axis=0)
    array([1, 1])
    >>> bn.nanargmax(a, axis=1)
    array([1, 0])
    
    """
    func, arr = nanargmin_selector(arr, axis)
    return func(arr)

def nanargmin_selector(arr, axis):
    """
    Return nanargmin function and array that matches `arr` and `axis`.
    
    Under the hood Bottleneck uses a separate Cython function for each
    combination of ndim, dtype, and axis. A lot of the overhead in
    bn.nanargmin() is in checking that `axis` is within range, converting
    `arr` into an array (if it is not already an array), and selecting the
    function to use to find the indices of the minimum.

    You can get rid of the overhead by doing all this before you, for example,
    enter an inner loop, by using the this function.

    Parameters
    ----------
    arr : array_like
        Input array. If `arr` is not an array, a conversion is attempted.
    axis : {int, None}
        Axis along which the indices are found.
    
    Returns
    -------
    func : function
        The nanargmin function that matches the number of dimensions and
        dtype of the input array and the axis.
    a : ndarray
        If the input array `arr` is not a ndarray, then `a` will contain the
        result of converting `arr` into a ndarray.

    Examples
    --------
    Create a numpy array:

    >>> arr = np.array([1.0, 2.0, 3.0])
    
    Obtain the function needed to determine the nanargmin of `arr` along
    axis=0:

    >>> func, a = bn.func.nanargmin_selector(arr, axis=0)
    >>> func
    <function nanargmin_1d_float64_axis0>
    
    Use the returned function and array to determine the maximum:
    
    >>> func(a)
    0

    """
    cdef np.ndarray a
    if type(arr) is np.ndarray:
        a = arr
    else:    
        a = np.array(arr, copy=False)
    cdef int ndim = PyArray_NDIM(a)
    cdef int dtype = PyArray_TYPE(a)
    if axis is not None:
        if axis < 0:
            axis += ndim
    else:
        a = PyArray_Ravel(a, NPY_CORDER)
        axis = 0
        ndim = 1
    cdef tuple key = (ndim, dtype, axis)
    try:
        func = nanargmin_dict[key]
    except KeyError:
        if (axis < 0) or (axis >= ndim):
            raise ValueError("axis(=%d) out of bounds" % axis)
        try:
            func = nanargmin_slow_dict[axis]
        except KeyError:
            tup = (str(ndim), str(a.dtype), str(axis))
            raise TypeError("Unsupported ndim/dtype/axis (%s/%s/%s)." % tup)
    return func, a
'''
